#include "vtkOSGRenderer.h"

#include "vtkCellArray.h"
#include "vtkCellData.h"
#include "vtkPointData.h"
#include "vtkPolyDataNormals.h"
#include "vtkPolyData.h"
#include "vtkPolyDataMapper.h"
#include "vtkProperty.h"
#include "vtkActor.h"

#include "vtkObjectFactory.h"

#include <osg/Geode>
#include <osg/Geometry>
#include <osgDB/WriteFile>
#include <osgProducer/Viewer>

vtkCxxRevisionMacro(vtkOSGRenderer, "$Revision: 1.2 $");
vtkStandardNewMacro(vtkOSGRenderer);

/*
	ViewerThread is a Singleton that opens a OpenProducer window and views all PolyDatas
	attached to an OSGRenderer.
*/
class ViewerThread : public OpenThreads::Thread, public OpenThreads::Mutex
{
	public:
		osgProducer::Viewer& getViewer() { return viewer; } 
		const osgProducer::Viewer& getViewer() const { return viewer; } 

		// Inserts new nodes into the graph - or replaces them if they have the same name...
		void addOrReplaceNode(osg::Node* node)
		{
			lock();
			_root->addChild(node);
			unlock();
		}

		static ViewerThread* instance()
		{
			static ViewerThread* _self = NULL;
			if(_self==NULL)
				_self = new ViewerThread;

			return _self;
		}

	protected:
		void run()
		{
			// set up the value with sensible default event handlers.
			viewer.setUpViewer(osgProducer::Viewer::STANDARD_SETTINGS);

			// create the windows and run the threads.
			viewer.realize();

			while( !viewer.done() )
			{
				// wait for all cull and draw threads to complete.
				viewer.sync();

				// update the scene by traversing it with the the update visitor which will
				// call all node update callbacks and animations.
				viewer.update();
		         
				// fire off the cull and draw traversals of the scene.
				viewer.frame();
		        
			}
		    
			// wait for all cull and draw threads to complete before exit.
			viewer.sync();
		}

		osgProducer::Viewer viewer;

	private:
		osg::ref_ptr<osg::Group> _root;

		ViewerThread()
		{
			_root = new osg::Group;
			viewer.setSceneData(_root.get());
		}
};


osg::Geometry *processPrimitive(vtkActor *actor, vtkCellArray *primArray, int primType)
{
	// get polyData from vtkActor
	vtkPolyData *polyData = (vtkPolyData *) actor->GetMapper()->GetInput();

	int numPrimitives = primArray->GetNumberOfCells();
	if (numPrimitives == 0) 
		return NULL;

	//Initialize the Geometry
	osg::Geometry* geom = new osg::Geometry;

	// get number of indices in the vtk prim array. Each vtkCell has the length
	// (not counted), followed by the indices.
	int primArraySize = primArray->GetNumberOfConnectivityEntries();
	int numIndices = primArraySize - numPrimitives;

	// allocate as many verts as there are indices in vtk prim array
	osg::Vec3Array* vertices = new osg::Vec3Array;

	// check to see if there are normals
	int normalPerVertex = 0;
	int normalPerCell = 0;
	vtkDataArray *normals = polyData->GetPointData()->GetNormals();
	if (actor->GetProperty()->GetInterpolation() == VTK_FLAT)
		normals = NULL;
	if (normals != NULL)
		normalPerVertex = 1;
	else
	{
		normals = polyData->GetCellData()->GetNormals();
		if (normals != NULL)
			normalPerCell = 1;
	}

	osg::Vec3Array* norms = new osg::Vec3Array;

	// check to see if there is color information
	int colorPerVertex = 0;
	int colorPerCell = 0;
	vtkUnsignedCharArray *colorArray = actor->GetMapper()->MapScalars(1.0);
	if (actor->GetMapper()->GetScalarVisibility() && colorArray != NULL)
	{
		int scalarMode = actor->GetMapper()->GetScalarMode();
		if (scalarMode == VTK_SCALAR_MODE_USE_CELL_DATA ||
			!polyData->GetPointData()->GetScalars()) // there is no point data
			colorPerCell = 1;
		else
			colorPerVertex = 1;
	}

	osg::Vec4Array* colors = new osg::Vec4Array;

	// check to see if there are texture coordinates
	vtkDataArray *texCoords = polyData->GetPointData()->GetTCoords();
	osg::Vec2Array* tcoords = new osg::Vec2Array;

	// copy data from vtk prim array to osg Geometry
	int prim = 0, vert = 0;
	int i, npts, totpts = 0, *pts;
	// go through cells (primitives)
	for (primArray->InitTraversal(); primArray->GetNextCell(npts, pts); prim++)
	{ 
		geom->addPrimitiveSet(new osg::DrawArrays(primType, totpts, npts));
		totpts += npts;
		if (colorPerCell)
		{
			unsigned char *aColor = colorArray->GetPointer(4*prim);
	        colors->push_back(osg::Vec4(aColor[0]/255.0f, aColor[1]/255.0f,
										aColor[2]/255.0f, aColor[3]/255.0f));
		}
		if (normalPerCell)
		{
			double *aNormal = normals->GetTuple(prim);
			norms->push_back(osg::Vec3(aNormal[0], aNormal[1], aNormal[2]));
		}
		// go through points in cell (verts)
		for (i=0; i < npts; i++)
		{
			double *aVertex = polyData->GetPoint(pts[i]);
			vertices->push_back(osg::Vec3(aVertex[0], aVertex[1], aVertex[2]));
			if (normalPerVertex)
			{
				double *aNormal = normals->GetTuple(pts[i]);
				norms->push_back(osg::Vec3(aNormal[0], aNormal[1], aNormal[2]));
			}
			if (colorPerVertex)
			{  
				unsigned char *aColor = colorArray->GetPointer(4*pts[i]);
				colors->push_back(osg::Vec4(aColor[0]/255.0f, aColor[1]/255.0f,
											aColor[2]/255.0f, aColor[3]/255.0f));
			}
			if (texCoords != NULL)
			{
				double *aTCoord = texCoords->GetTuple(pts[i]);
				tcoords->push_back(osg::Vec2(aTCoord[0], aTCoord[1]));
			}
			vert++;
		}
	}

	// add attribute arrays to gset
	geom->setVertexArray(vertices);
    geom->setColorArray(colors);
    if(normals) geom->setNormalArray(norms);

	if (normalPerVertex)
        geom->setNormalBinding(osg::Geometry::BIND_PER_VERTEX);
	if (normalPerCell)
        geom->setNormalBinding(osg::Geometry::BIND_PER_PRIMITIVE);

	if (colorPerVertex)
        geom->setColorBinding(osg::Geometry::BIND_PER_VERTEX);
	else if (colorPerCell)
        geom->setColorBinding(osg::Geometry::BIND_PER_PRIMITIVE);
	else
	{ 
    // use overall color (get from Actor)
		double *actorColor = actor->GetProperty()->GetColor();
		float opacity = actor->GetProperty()->GetOpacity();

		colors->push_back(osg::Vec4(actorColor[0], actorColor[1], actorColor[2], opacity));
        geom->setColorBinding(osg::Geometry::BIND_OVERALL);
	}
  
	if (texCoords != NULL)
		geom->setTexCoordArray(0,tcoords);


	// create a geostate for this geoset
	osg::StateSet* stateset = new osg::StateSet;

	// if not opaque
	if (actor->GetProperty()->GetOpacity() < 1.0)
	{
		stateset->setRenderingHint(osg::StateSet::TRANSPARENT_BIN); // draw last
	    stateset->setMode(GL_LIGHTING,osg::StateAttribute::OFF);
		stateset->setMode(GL_BLEND,osg::StateAttribute::ON);
		stateset->setMode(GL_CULL_FACE,osg::StateAttribute::OFF);
	}
/*
	// wireframe - how do I set this ?
	if (actor->GetProperty()->GetRepresentation() == VTK_WIREFRAME) 
		stateset->setMode(GL_WIREFRAME, osg::StateAttribute::ON);
*/
	// backface culling
	if (!actor->GetProperty()->GetBackfaceCulling())
		stateset->setMode(GL_CULL_FACE,osg::StateAttribute::OFF);
  
	// lighting
	if (normals != NULL)
	    stateset->setMode(GL_LIGHTING,osg::StateAttribute::ON);
	else
	    stateset->setMode(GL_LIGHTING,osg::StateAttribute::OFF);

	// if it is lines, turn off lighting. 
	if (primType == osg::PrimitiveSet::LINE_STRIP)
	    stateset->setMode(GL_LIGHTING,osg::StateAttribute::OFF);

    geom->setStateSet(stateset);
	return geom;
}

//----------------------------------------------------------------------------
vtkOSGRenderer::vtkOSGRenderer()
{
}

//----------------------------------------------------------------------------
vtkOSGRenderer::~vtkOSGRenderer()
{
}

//----------------------------------------------------------------------------
void vtkOSGRenderer::Execute()
{
	osg::ref_ptr<osg::Geode> geode = new osg::Geode();

	ViewerThread* vthread = ViewerThread::instance();

	if(!vthread->isRunning())
		vthread->start();

	vtkPolyData *input = this->GetInput();
	vtkPolyData *output = this->GetOutput();

	// this is to make a dummy forward to ParaViews render
	output->SetVerts(input->GetVerts());
	output->SetLines(input->GetLines());
	output->SetPolys(input->GetPolys());
	output->SetStrips(input->GetStrips());
	output->SetPoints(input->GetPoints());
// remember the colors !!!

	vtkPolyDataNormals* normals = vtkPolyDataNormals::New();
    normals->SetInput( input );
	normals->Update();
	//
	// We create an instance of vtkPolyDataMapper to map the polygonal data 
	// into graphics primitives. We connect the output of the cone souece 
	// to the input of this mapper 
	//
	vtkPolyDataMapper *mapper = vtkPolyDataMapper::New();
	mapper->SetInput( normals->GetOutput() );

	//
	// create an actor to represent the cone. The actor coordinates rendering of
	// the graphics primitives for a mapper. We set this actor's mapper to be
	// coneMapper which we created above.
	//
	vtkActor *actor = vtkActor::New();
	actor->SetMapper( mapper );


	// get primitive arrays
	osg::Geometry* points, *lines, *polys, *strips;

	// create new Geometry for the Geode
	points = processPrimitive(actor, input->GetVerts(), osg::PrimitiveSet::POINTS);
	lines = processPrimitive(actor, input->GetLines(), osg::PrimitiveSet::LINE_STRIP);
	polys = processPrimitive(actor, input->GetPolys(), osg::PrimitiveSet::POLYGON);
	strips = processPrimitive(actor, input->GetStrips(), osg::PrimitiveSet::TRIANGLE_STRIP);

	// remove old gsets and delete them
	while( geode->getNumDrawables() ) geode->removeDrawable((int)0);

	if( points ) geode->addDrawable( points );
	if( lines ) geode->addDrawable( lines );
	if( polys ) geode->addDrawable( polys );
	if( strips ) geode->addDrawable( strips );

	vthread->addOrReplaceNode(geode.get());
//	vthread.lock();
//	vthread.getViewer().setSceneData(geode.get());
//	vthread.unlock();
}


//----------------------------------------------------------------------------
void vtkOSGRenderer::PrintSelf(ostream& os, vtkIndent indent)
{
  this->Superclass::PrintSelf(os,indent);
}
